package com.fg.scripting.groovy.classloader;

import com.fg.scripting.*;
import com.fg.scripting.exception.*;
import com.fg.scripting.groovy.aop.*;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.BeanWrapper;
import org.springframework.beans.BeanWrapperImpl;
import org.springframework.beans.BeansException;
import org.springframework.beans.MutablePropertyValues;
import org.springframework.beans.PropertyAccessor;
import org.springframework.beans.PropertyValue;
import org.springframework.beans.PropertyValues;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.support.DefaultListableBeanFactory;
import org.springframework.beans.factory.support.InstantiationStrategy;
import org.springframework.beans.factory.support.RootBeanDefinition;
import org.springframework.beans.factory.support.SimpleInstantiationStrategy;

import java.lang.reflect.Constructor;

/**
 * Description
 *
 * @author Jan Novotný, FG Forrest a.s. (c) 2007
 * @version $Id: ScriptingBeanFactory.java,v 1.9 2010/05/03 08:58:23 u_novoj Exp $
 */
public class ScriptingBeanFactory extends DefaultListableBeanFactory {
	private static final Log log = LogFactory.getLog(ScriptingBeanFactory.class);
	private final ScriptingFactory scriptingFactory;
	private InstantiationStrategy wrappingInstantiationStrategy;

	public ScriptingBeanFactory(ScriptingFactory scriptingFactory) {
		this.scriptingFactory = scriptingFactory;
	}

	public ScriptingBeanFactory(BeanFactory parentBeanFactory, ScriptingFactory scriptingFactory) {
		super(parentBeanFactory);
		this.scriptingFactory = scriptingFactory;
	}

	@Override
	protected InstantiationStrategy getInstantiationStrategy() {
		if(wrappingInstantiationStrategy == null) {
			wrappingInstantiationStrategy = new ScriptingFactoryInstantiationStrategy(
					scriptingFactory,
					super.getInstantiationStrategy()
			);
		}
		return wrappingInstantiationStrategy;
	}

	/**
	 * For scripting proxies store wired property values into the initializer, so that initializer can take care of
	 * proper initialization of newly refreshed instances.
	 * @param beanName
	 * @param mbd
	 * @param bw
	 * @param pvs
	 */
	@Override
	protected void applyPropertyValues(String beanName, BeanDefinition mbd, BeanWrapper bw, PropertyValues pvs) {
		Object beanInstance = bw.getWrappedInstance();
		if(beanInstance instanceof ScriptingProxy) {
			ScriptingProxy scriptingProxy = (ScriptingProxy)beanInstance;
			SpringProxyInitializer initializer = (SpringProxyInitializer)scriptingProxy.getInstanceInitializer();
			MonitoredBeanWrapper mbw = new MonitoredBeanWrapper(scriptingProxy.unwrap());
			super.applyPropertyValues(beanName, mbd, mbw, pvs);
			initializer.setPropertyValues(mbw.getPvs());
		} else {
			super.applyPropertyValues(beanName, mbd, bw, pvs);
		}
	}

	/**
	 * For scripted classes creates initializer instance, that is able to recreate and reinitialize newly created proxy.
	 */
	private class ScriptingFactoryInstantiationStrategy extends SimpleInstantiationStrategy {
		private final ScriptingFactory scriptingFactory;
		private final InstantiationStrategy defaultStrategy;

		private ScriptingFactoryInstantiationStrategy(ScriptingFactory scriptingFactory, InstantiationStrategy defaultStrategy) {
			this.scriptingFactory = scriptingFactory;
			this.defaultStrategy = defaultStrategy;
		}

		public Object instantiate(RootBeanDefinition beanDefinition, String beanName, BeanFactory owner) {
			if(isClassLoadedByScriptingClassloader(beanDefinition.getBeanClass())) {
				Object scriptingInstance = scriptingFactory.createInstance(beanDefinition.getBeanClass());
				if (scriptingInstance instanceof ScriptingProxy) {
					ScriptingProxy proxy = (ScriptingProxy)scriptingInstance;
					proxy.setInstanceInitializer(new SpringProxyInitializer(beanName));
					return proxy;
				} else {
					return scriptingInstance;
				}
			} else {
				return defaultStrategy.instantiate(beanDefinition, beanName, owner);
			}
		}

		public Object instantiate(RootBeanDefinition beanDefinition, String beanName, BeanFactory owner, Constructor ctor, Object[] args) {
			if(isClassLoadedByScriptingClassloader(beanDefinition.getBeanClass())) {
				Object scriptingInstance = scriptingFactory.createInstance(beanDefinition.getBeanClass(), args);
				if (scriptingInstance instanceof ScriptingProxy) {
					ScriptingProxy proxy = (ScriptingProxy)scriptingInstance;
					proxy.setInstanceInitializer(new SpringProxyInitializer(beanName, args));
					return proxy;
				} else {
					return scriptingInstance;
				}
			} else {
				return defaultStrategy.instantiate(beanDefinition, beanName, owner, ctor, args);
			}
		}

		/**
		 * Returns true if class is loaded by ScriptingClassloader.
		 *
		 * @param clazz
		 * @return
		 */
		private boolean isClassLoadedByScriptingClassloader(Class clazz) {
			ClassLoader parameterClassLoader = clazz.getClassLoader();
			do {
				if(parameterClassLoader == scriptingFactory.getScriptingClassLoader()) {
					return true;
				}
				parameterClassLoader = parameterClassLoader.getParent();
			} while(parameterClassLoader != ClassLoader.getSystemClassLoader());
			return false;
		}

	}

	/**
	 * Recreates instance of the scripting class with proper constructor and performs proper wiring logic and
	 * after initialization Spring lifecycle callbacks. 
	 */
	private class SpringProxyInitializer extends GroovyProxyMixin.DefaultProxyInstanceFactory {
		private final Object[] constructorArgs;
		private final Class[] constructorArgTypes;
		private final String beanName;
		private PropertyValues propertyValues;

		private SpringProxyInitializer(String beanName, Object... constructorArgs) {
			this.beanName = beanName;
			this.constructorArgs = constructorArgs;
			this.constructorArgTypes = new Class[constructorArgs.length];
			for(int i = 0; i < constructorArgs.length; i++) {
				Object constructorArg = constructorArgs[i];
				this.constructorArgTypes[i] = constructorArg.getClass();
			}
		}

		public void setPropertyValues(PropertyValues propertyValues) {
			this.propertyValues = propertyValues;
		}

		public Object createInstance(Class clazz) throws ClassInstantiationException {
			try {
				Object o;
				if(constructorArgTypes.length > 0) {
					Constructor constructor = clazz.getDeclaredConstructor(constructorArgTypes);
					o = constructor.newInstance(constructorArgs);
				} else {
					o = clazz.newInstance();
				}
				if(propertyValues != null) {
					if(!propertyValues.isEmpty()) {
						PropertyAccessor wrapper = new MonitoredBeanWrapper(o);
						wrapper.setPropertyValues(propertyValues);
					}
				} else {
					if(log.isWarnEnabled()) {
						log.warn("No property values initialized for sripting instance (" + o.getClass().getName() + ") refresh!");
					}
				}
				ScriptingBeanFactory.this.initializeBean(o, beanName);
				return o;
			}
			catch(Exception e) {
				String msg = "Cannot reinstantiate class: " + clazz.getName() + " (" + e.getLocalizedMessage() + ')';
				log.error(msg);
				throw new ClassInstantiationException(msg, e, clazz);
			}
		}

	}

	/**
	 * Monitors properties set throught it.
	 */
	private class MonitoredBeanWrapper extends BeanWrapperImpl {
		private PropertyValues pvs;

		private MonitoredBeanWrapper(Object object) {
			super(object);
		}

		public PropertyValues getPvs() {
			return pvs;
		}

		@Override
		public void setPropertyValues(PropertyValues pvs, boolean ignoreUnknown, boolean ignoreInvalid) throws BeansException {
			MutablePropertyValues mpvs = new MutablePropertyValues();
			for(PropertyValue value : pvs.getPropertyValues()) {
				Object valueObject = value.getValue();
				if (valueObject instanceof ScriptingProxy) {
					mpvs.addPropertyValue(value.getName(), ((ScriptingProxy)valueObject).unwrap());
				} else {
					mpvs.addPropertyValue(value);
				}
			}
			super.setPropertyValues(mpvs, ignoreUnknown, ignoreInvalid);
			this.pvs = pvs;
		}

	}

}