package com.fg.scripting.groovy.classloader;

import com.fg.scripting.exception.*;
import groovy.lang.GroovyResourceLoader;
import groovy.util.ResourceConnector;
import groovy.util.ResourceException;
import org.springframework.core.io.Resource;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;

/**
 * Represents bridge to Groovy for loading resources from Spring resource loader.
 *
 * @author Jan Novotný, FG Forrest a.s. (c) 2007
 * @version $Id: SimpleGroovyResourceLoader.java,v 1.4 2010/01/11 15:46:16 u_novoj Exp $
 */
public class SimpleGroovyResourceLoader implements ResourceConnector, GroovyResourceLoader {
	private final Resource[] baseDirectories;
	private final String groovyDirectoriesDescription;

	public SimpleGroovyResourceLoader(Resource baseDir) {
		if (baseDir == null) throw new IllegalArgumentException("Non null base directory for Groovy scripts required!");
		this.baseDirectories = new Resource[] {baseDir};
		this.groovyDirectoriesDescription = createDescription();
	}

	public SimpleGroovyResourceLoader(Resource[] baseDirectories) {
		if (baseDirectories == null || baseDirectories.length == 0) {
			throw new IllegalArgumentException("Non null base directory for Groovy scripts required!");
		}
		this.baseDirectories = baseDirectories;
		this.groovyDirectoriesDescription = createDescription();
	}	

	private String createDescription() {
		StringBuffer description = new StringBuffer();
		for(Resource baseDirectory : baseDirectories) {
			if (description.length() > 0) {
				description.append(", ");
			}
			description.append(baseDirectory.getDescription());
		}
		return description.toString();
	}

	/**
	 * Retrieve a URLConnection to a script referenced by name.
	 *
	 * @param name
	 * @throws groovy.util.ResourceException
	 */
	public URLConnection getResourceConnection(String name) throws ResourceException {
		if (!name.endsWith(".groovy")) {
			name = name.replace('.', File.separatorChar) + ".groovy";
		}
		try {
			Resource groovyResource;
			for(Resource baseDirectory : baseDirectories) {
				groovyResource = baseDirectory.createRelative(name);
				if (groovyResource.exists()) {
					return groovyResource.getURL().openConnection();
				}
			}
			throw new ResourceException("Source " + name + " doesn't exists in any of Groovy source directories: " + groovyDirectoriesDescription + "!");
		} catch (IOException ex) {
			String msg = "Groovy class " + name + " doesn't exist!";
			throw new ResourceException(msg, ex);
		}
	}

	/**
	 * Loads a Groovy source file given its name.
	 *
	 * @param filename name of the file
	 * @return a URL
	 * @throws java.net.MalformedURLException if the URL is invalid
	 */
	public URL loadGroovySource(String filename) throws MalformedURLException {
		if (!filename.endsWith(".groovy")) {
			filename = filename.replace('.', File.separatorChar) + ".groovy";
		}
		try {
			Resource groovyResource;
			for(Resource baseDirectory : baseDirectories) {
				groovyResource = baseDirectory.createRelative(filename);
				if (groovyResource.exists()) {
					return groovyResource.getURL();
				}
			}
			return null;
		} catch (IOException ex) {
			String msg = "Groovy class " + filename + " doesn't exist!";
			throw new ClassLoadingException(msg, ex);
		}
	}

}
